use crate::{slice2d::Slice2D, Element, ImageData};
use bitvec::vec::BitVec;

pub fn read_element_bits(input: &mut dyn Iterator<Item = bool>) -> BitVec {
    let mut output = BitVec::new();
    loop {
        output.extend(input.take(7));
        if !input.next().unwrap_or_default() {
            break;
        }
    }
    output
}

pub fn read_quad_tree(input: &mut dyn Iterator<Item = bool>, output: Slice2D) -> Result<(), ()> {
    let header = input.take(4).collect::<Vec<bool>>();
    if header.len() != 4 {
        return Err(());
    }
    let (hlen, vlen) = output.len();
    let hmid = hlen / 2;
    let vmid = vlen / 2;
    let quadrants = [
        output.get_slice2d((0..hmid, 0..vmid)),
        output.get_slice2d((hmid..hlen, 0..vmid)),
        output.get_slice2d((0..hmid, vmid..vlen)),
        output.get_slice2d((hmid..hlen, vmid..vlen)),
    ];
    if hlen == 2 || vlen == 2 {
        for (quadrant, v) in quadrants.into_iter().zip(header.into_iter()) {
            quadrant.set(v, (0, 0));
        }
    } else {
        for (quadrant, heterogenous) in quadrants.into_iter().zip(header.into_iter()) {
            if heterogenous {
                if let Some(value) = input.next() {
                    quadrant.set_all(value);
                } else {
                    return Err(());
                }
            } else {
                read_quad_tree(input, quadrant)?;
            }
        }
    }
    Ok(())
}

pub fn parse_element(input: &mut dyn Iterator<Item = bool>) -> Result<Element, ()> {
    let element_bits = read_element_bits(input);
    let mut input = element_bits.into_iter();
    if let Some(is_glyph) = input.next() {
        if is_glyph {
            if let Some(cs) = input.next() {
                let image_data = ImageData::default();
                let slice_2d = image_data.get_slice2d((0..64, 0..64));
                read_quad_tree(&mut input, slice_2d)?;
                Ok(Element::Glyph {
                    courtesy_space: cs,
                    image_data: Box::new(image_data),
                })
            } else {
                Err((/* Not enough data! */))
            }
        } else {
            Ok(Element::ControlCode {
                data: input
                    .take(128)
                    .enumerate()
                    .fold(0, |a, (i, b)| a | ((b as u128) << i)),
            })
        }
    } else {
        Err((/* No data! */))
    }
}

pub fn decode(input: &[u8]) -> Result<Vec<Element>, ()> {
    let mut output = Vec::new();
    let bits = BitVec::<u8>::from_slice(input);
    let mut iter = bits.into_iter().peekable();
    while iter.peek().is_some() {
        output.push(parse_element(&mut iter)?);
    }
    Ok(output)
}
