use crate::{slice2d::Slice2D, Element};
use bitvec::{slice::BitSlice, vec::BitVec};

pub fn write_element_bits(bits: &BitSlice<u8>, output: &mut BitVec<u8>) {
    let mut iter = bits.chunks(7).peekable();
    while let Some(next_7) = iter.next() {
        if iter.peek().is_some() {
            output.push(true);
        } else {
            output.push(false);
        }
        output.extend_from_bitslice(next_7);
    }
}

pub fn write_quad_tree(input: Slice2D, output: &mut BitVec<u8>) {
    let (hlen, vlen) = input.len();
    let hmid = hlen / 2;
    let vmid = vlen / 2;
    let quadrants = [
        input.get_slice2d((0..hmid, 0..vmid)),
        input.get_slice2d((hmid..hlen, 0..vmid)),
        input.get_slice2d((0..hmid, vmid..vlen)),
        input.get_slice2d((hmid..hlen, vmid..vlen)),
    ];
    let mut buffer = BitVec::<u8>::new();
    if hlen == 2 || vlen == 2 {
        for quadrant in quadrants {
            buffer.push(quadrant.get((0, 0)));
        }
    } else {
        for quadrant in quadrants {
            if let Some(v) = quadrant.get_all() {
                output.push(false);
                buffer.push(v);
            } else {
                output.push(true);
                write_quad_tree(quadrant, &mut buffer);
            }
        }
    }
    output.extend_from_bitslice(&buffer);
}

pub fn write_element(input: Element, output: &mut BitVec<u8>) {
    match input {
        Element::ControlCode { data } => {
            let mut bits = BitVec::<u8>::new();
            bits.push(false);
            bits.extend_from_raw_slice(&data.to_le().to_le_bytes());
            let mut bitslice = bits.as_bitslice();
            for (i, bit) in bits.iter().enumerate() {
                if *bit {
                    bitslice = &bits[i..];
                    break;
                }
            }
            write_element_bits(bitslice, output);
        }
        Element::Glyph {
            courtesy_space,
            image_data,
        } => {
            let mut bits = BitVec::<u8>::new();
            bits.push(true);
            bits.push(courtesy_space);
            for _ in 0..5 {
                bits.push(false);
            }
            write_quad_tree(image_data.get_slice2d((0..64, 0..64)), &mut bits);
            write_element_bits(&bits, output);
        }
    }
}

pub fn encode(input: Vec<Element>) -> Vec<u8> {
    let mut output = Vec::new();
    for element in input {
        let mut bits = BitVec::<u8>::new();
        write_element(element, &mut bits);
        bits.set_uninitialized(false);
        output.extend_from_slice(&bits.into_vec());
    }
    output
}
